<?php

namespace EventPortal\Domain\Util\IOEngage;

use Carbon\Carbon;
use EventPortal\Domain\Util\IOEngage\Exceptions\AuthenticationFailed;
use EventPortal\Domain\Util\IOEngage\Exceptions\DelegateRetrievalFailed;
use EventPortal\Domain\Util\IOEngage\Exceptions\SearchFailed;
use GuzzleHttp\Client;
use Illuminate\Support\Arr;
use Psr\Http\Message\ResponseInterface;

class IOEngage
{
    /**
     * @var string
     */
    protected $exception;

    /**
     * @var Client
     */
    protected $http;

    public function __construct()
    {
        $this->http = new Client();
    }

    // search
    public function search(array $args): object
    {
        $response = $this->makeAuthenticatedRequest(
            $this->buildUrlParams(
                config('ioengage.endpoints.search'),
                $args
            ),
            SearchFailed::class
        );

        return $this->getContents($response, [
            'PerPage',
            'CurrentPage',
            'TotalPages',
            'TotalResults',
            'RecordCount',
            'Results',
        ]);
    }

    // getDelegates

    public function delegates(string ...$oneKeyIds): object
    {
        $response = $this->makeAuthenticatedRequest(
            config('ioengage.endpoints.delegate') . implode(',', $oneKeyIds),
            DelegateRetrievalFailed::class,
        );

        return $this->getContents($response, ['RecordCount', 'Results']);
    }

    // markAttendance

    public function authenticate(): self
    {
        if ($this->needsToAuthenticate() === false) {
            return $this;
        }

        $this->exception = AuthenticationFailed::class;

        $url = $this->buildUrlParams(
            config('ioengage.endpoints.token'),
            config('ioengage.authentication'),
        );

        $response = $this->http->post($url);

        $payload = $this->getContents($response, ['access_token', 'expires_in']);

        cache(['ioengage' => [
            'token' => $payload->access_token,
            'expires' => Carbon::now()->addSeconds($payload->expires_in),
        ]]);

        return $this;
    }

    private function makeAuthenticatedRequest(string $uri, string $exception, string $method = 'get'): ResponseInterface
    {
        $this->authenticate();

        $this->exception = $exception;

        ['token' => $token] = cache('ioengage');

        return $this->http->$method($uri, ['headers' => ['Authorization' => "Bearer {$token}"]]);
    }

    private function getContents(ResponseInterface $response, array $keys, int $status = 200): object
    {
        throw_if(
            $response->getStatusCode() !== $status
                || $this->payloadContains($response, $keys),
            new $this->exception
        );

        return json_decode($response->getBody());
    }

    private function payloadContains(ResponseInterface $response, array $keys): bool
    {
        $decode = json_decode($response->getBody()->getContents(), true);

        return Arr::has(array_keys($decode), $keys);
    }

    /**
     * Checks the current cache for the existence of credentials, then checks the expiry to determine whether the token is still valid.
     */
    private function needsToAuthenticate(): bool
    {
        $cache = cache('ioengage');
        if ($cache !== null
            && Arr::get($cache, 'token') !== null
            && Arr::get($cache, 'expires') > Carbon::now()
        ) {
            return false;
        }

        return true;
    }

    private function buildUrlParams(string $baseUri, array $args): string
    {
        return "{$baseUri}?" . implode('&', collect($args)->map(function ($value, $key): string {
            return "{$key}={$value}";
        })->toArray());
    }
}
